#include "winver.h"

1 TYPELIB "multipleax.rc"
1 ICON "multipleax.ico"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION 1,0,0,0
 PRODUCTVERSION 1,0,0,0
 FILEFLAGSMASK 0x3fL
 FILEOS 0x00040000L
 FILETYPE 0x2L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904e4"
        BEGIN
            VALUE "CompanyName", "The Qt Company Ltd."
            VALUE "FileDescription", "Multiple Example (ActiveQt)"
            VALUE "FileVersion", "1.0.0.0"
            VALUE "LegalCopyright", "Copyright (C) 2015 The Qt Company Ltd."
            VALUE "InternalName", "multipleax.dll"
            VALUE "OriginalFilename", "multipleax.dll"
            VALUE "ProductName", "Multiple Example (ActiveQt)"
            VALUE "ProductVersion", "1.0.0.0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252
    END
END
